
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2020-12-22T17:24:22-0800",
 "dirty": false,
 "error": null,
 "full-revisionid": "f64f6c6e3847351b4c6f16285cd4fcb8a40e8f93",
 "version": "0.13.4"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
